import unittest

from transformers import LlamaConfig, LlamaForCausalLM, PreTrainedModel


def assert_model_equality(
    test_case: unittest.TestCase, model1: PreTrainedModel, model2: PreTrainedModel
):
    """Checks if two models are equal."""
    test_case.assertEqual(type(model1), type(model2))
    test_case.assertEqual(str(model1.state_dict()), str(model2.state_dict()))


def get_test_model() -> PreTrainedModel:
    """Gets a test model that is small enough to load and store quickly.

    Returns:
        PreTrainedModel: Tiny test model.
    """
    return LlamaForCausalLM(
        LlamaConfig(
            vocab_size=2000,
            hidden_size=256,
            intermediate_size=688,
            num_hidden_layers=2,
            num_attention_heads=2,
        )
    )



import time
import unittest

from taoverse.utilities.perf_monitor import PerfMonitor


class TestPerfMonitor(unittest.TestCase):
    def test_format_duration(self):
        """Performs basic validation of the _format_duration method."""

        monitor = PerfMonitor("test")

        self.assertEqual(monitor._format_duration(1), "1.00 ns")
        self.assertEqual(monitor._format_duration(2500), "2.50 μs")
        self.assertEqual(monitor._format_duration(3000000), "3.00 ms")
        self.assertEqual(monitor._format_duration(1230000000), "1.23 s")
        self.assertEqual(monitor._format_duration(120000000000), "2.00 min")

    def test_perf_monitor(self):
        """Performs basic validation of the PerfTracker and its output_str."""

        tracker = PerfMonitor("TestOfTime")

        self.assertRegex(tracker.summary_str(), "TestOfTime performance: N=0")

        with tracker.sample():
            time.sleep(1)

        self.assertRegex(
            tracker.summary_str(),
            r"TestOfTime performance: N=1 \| Min=1.[0-9]{2} s \| Max=1.[0-9]{2} s \| Median=1.[0-9]{2} s \| P90=1.[0-9]{2} s",
        )

        with tracker.sample():
            time.sleep(4)

        self.assertRegex(
            tracker.summary_str(),
            r"TestOfTime performance: N=2 \| Min=1.[0-9]{2} s \| Max=4.[0-9]{2} s \| Median=2.[0-9]{2} s \| P90=3.[0-9]{2} s",
        )

    def test_perf_monitor_min(self):
        """Performs basic validation of the min method."""

        tracker = PerfMonitor("TestMin")
        tracker.set_samples_for_testing([1_000_000_000, 4_000_000_000])

        self.assertEqual(tracker.min(), 1.0)

    def test_perf_monitor_max(self):
        """Performs basic validation of the max method."""
        tracker = PerfMonitor("TestMax")
        tracker.set_samples_for_testing([1_000_000_000, 4_000_000_000])
        self.assertEqual(tracker.max(), 4.0)

    def test_perf_monitor_median(self):
        """Performs basic validation of the median method."""
        tracker = PerfMonitor("TestMedian")
        tracker.set_samples_for_testing([1_000_000_000, 2_000_000_000, 3_000_000_000])
        self.assertEqual(tracker.median(), 2.0)

    def test_perf_monitor_percentile(self):
        """Performs basic validation of the percentile method."""
        tracker = PerfMonitor("TestPercentile")
        tracker.set_samples_for_testing(
            [1_000_000_000, 2_000_000_000, 3_000_000_000, 4_000_000_000]
        )
        self.assertEqual(tracker.percentile(90), 3.7)


if __name__ == "__main__":
    unittest.main()



from dataclasses import replace
import os
import unittest

from taoverse.model.data import ModelId, ModelMetadata
from taoverse.model.model_tracker import ModelTracker


class TestModelTracker(unittest.TestCase):
    def setUp(self):
        self.model_tracker = ModelTracker()

    def _create_model_id(self) -> ModelId:
        return ModelId(
            namespace="test_model",
            name="test_name",
            commit="test_commit",
            hash="test_hash",
            secure_hash="test_secure_hash",
            competition_id=1,
        )

    def test_roundtrip_state(self):
        hotkey = "test_hotkey"
        model_id = self._create_model_id()
        model_metadata = ModelMetadata(id=model_id, block=1)

        state_path = ".test_tracker_state.pickle"
        self.model_tracker.on_miner_model_updated(hotkey, model_metadata)
        self.model_tracker.save_state(state_path)

        new_tracker = ModelTracker()
        new_tracker.load_state(state_path)

        os.remove(state_path)

        self.assertEqual(
            self.model_tracker.miner_hotkey_to_model_metadata_dict,
            new_tracker.miner_hotkey_to_model_metadata_dict,
        )

    def test_on_miner_model_updated_add(self):
        hotkey = "test_hotkey"
        model_id = self._create_model_id()

        model_metadata = ModelMetadata(id=model_id, block=1)

        self.model_tracker.on_miner_model_updated(hotkey, model_metadata)

        self.assertTrue(
            hotkey in self.model_tracker.miner_hotkey_to_model_metadata_dict
        )
        self.assertEqual(
            model_metadata,
            self.model_tracker.miner_hotkey_to_model_metadata_dict[hotkey],
        )

    def test_on_miner_model_updated_update(self):
        hotkey = "test_hotkey"
        model_id = self._create_model_id()

        model_metadata = ModelMetadata(id=model_id, block=1)

        new_model_id = replace(
            model_id,
            commit="other_commit",
            hash="other_hash",
            secure_hash="other_secure_hash",
        )
        new_model_metadata = ModelMetadata(id=new_model_id, block=2)

        self.model_tracker.on_miner_model_updated(hotkey, model_metadata)
        self.model_tracker.on_miner_model_updated(hotkey, new_model_metadata)

        self.assertTrue(
            hotkey in self.model_tracker.miner_hotkey_to_model_metadata_dict
        )
        self.assertEqual(
            new_model_metadata,
            self.model_tracker.miner_hotkey_to_model_metadata_dict[hotkey],
        )

    def test_get_model_metadata_for_miner_hotkey(self):
        hotkey = "test_hotkey"
        model_id = model_id = self._create_model_id()

        model_metadata = ModelMetadata(id=model_id, block=1)

        self.model_tracker.on_miner_model_updated(hotkey, model_metadata)
        returned_model_metadata = (
            self.model_tracker.get_model_metadata_for_miner_hotkey(hotkey)
        )

        self.assertEqual(model_metadata, returned_model_metadata)

    def test_get_model_metadata_for_miner_hotkey_optional(self):
        hotkey = "test_hotkey"

        returned_model_id = self.model_tracker.get_model_metadata_for_miner_hotkey(
            hotkey
        )

        self.assertIsNone(returned_model_id)

    def test_get_miner_hotkey_to_model_metadata_dict(self):
        hotkey_1 = "test_hotkey"
        model_id_1 = ModelId(
            namespace="test_model",
            name="test_name",
            commit="test_commit",
            hash="test_hash",
            secure_hash="test_secure_hash",
            competition_id=1,
        )
        model_metadata_1 = ModelMetadata(id=model_id_1, block=1)

        hotkey_2 = "test_hotkey2"
        model_id_2 = ModelId(
            namespace="test_model2",
            name="test_name2",
            commit="test_commit2",
            hash="test_hash2",
            secure_hash="test_secure_hash2",
            competition_id=1,
        )
        model_metadata_2 = ModelMetadata(id=model_id_2, block=2)

        self.model_tracker.on_miner_model_updated(hotkey_1, model_metadata_1)
        self.model_tracker.on_miner_model_updated(hotkey_2, model_metadata_2)

        hotkey_to_model_metadata = (
            self.model_tracker.get_miner_hotkey_to_model_metadata_dict()
        )

        self.assertEqual(len(hotkey_to_model_metadata), 2)
        self.assertEqual(hotkey_to_model_metadata[hotkey_1], model_metadata_1)
        self.assertEqual(hotkey_to_model_metadata[hotkey_2], model_metadata_2)

    def test_on_hotkeys_updated_extra_ignored(self):
        hotkey = "test_hotkey"
        model_id = self._create_model_id()

        model_metadata = ModelMetadata(id=model_id, block=1)

        self.model_tracker.on_miner_model_updated(hotkey, model_metadata)
        self.model_tracker.on_hotkeys_updated(set([hotkey, "extra_hotkey"]))

        self.assertEqual(len(self.model_tracker.miner_hotkey_to_model_metadata_dict), 1)

    def test_on_hotkeys_updated_missing_removed(self):
        hotkey = "test_hotkey"
        model_id = self._create_model_id()

        model_metadata = ModelMetadata(id=model_id, block=1)

        self.model_tracker.on_miner_model_updated(hotkey, model_metadata)
        self.model_tracker.on_hotkeys_updated(set(["extra_hotkey"]))

        self.assertEqual(len(self.model_tracker.miner_hotkey_to_model_metadata_dict), 0)


if __name__ == "__main__":
    unittest.main()



import base64
import hashlib
from typing import Tuple

from taoverse.model.data import ModelId, ModelMetadata

def validate_hf_repo_id(repo_id: str) -> Tuple[str, str]:
    """Verifies a Hugging Face repo id is valid and returns it split into namespace and name.

    Raises:
        ValueError: If the repo id is invalid.
    """

    if not repo_id:
        raise ValueError("Hugging Face repo id cannot be empty.")

    if not 3 < len(repo_id) <= ModelId.MAX_REPO_ID_LENGTH:
        raise ValueError(
            f"Hugging Face repo id must be between 3 and {ModelId.MAX_REPO_ID_LENGTH} characters."
        )

    parts = repo_id.split("/")
    if len(parts) != 2:
        raise ValueError(
            "Hugging Face repo id must be in the format <org or user name>/<repo_name>."
        )

    return parts[0], parts[1]


def get_hf_url(model_metadata: ModelMetadata) -> str:
    """Returns the URL to the Hugging Face repo for the provided model metadata."""
    return f"https://huggingface.co/{model_metadata.id.namespace}/{model_metadata.id.name}/tree/{model_metadata.id.commit}"


def get_hash_of_two_strings(string1: str, string2: str) -> str:
    """Hashes two strings together and returns the result."""

    string_hash = hashlib.sha256((string1 + string2).encode())

    return base64.b64encode(string_hash.digest()).decode("utf-8")



import os
import tempfile
from dataclasses import replace
from typing import Optional

from huggingface_hub import HfApi
from transformers import AutoModelForCausalLM

from taoverse.model.competition.data import ModelConstraints
from taoverse.model.data import Model, ModelId
from taoverse.model.storage.disk import utils
from taoverse.model.storage.remote_model_store import RemoteModelStore


class HuggingFaceModelStore(RemoteModelStore):
    """Hugging Face based implementation for storing and retrieving a model."""

    @classmethod
    def assert_access_token_exists(cls) -> str:
        """Asserts that the access token exists."""
        if not os.getenv("HF_ACCESS_TOKEN"):
            raise ValueError("No Hugging Face access token found to write to the hub.")
        return os.getenv("HF_ACCESS_TOKEN")

    @classmethod
    def get_access_token_if_exists(cls) -> Optional[str]:
        """Returns the access token if it exists."""
        return os.getenv("HF_ACCESS_TOKEN")

    async def upload_model(
        self, model: Model, model_constraints: ModelConstraints
    ) -> ModelId:
        """Uploads a trained model to Hugging Face."""
        token = HuggingFaceModelStore.assert_access_token_exists()

        commit_info = model.pt_model.push_to_hub(
            repo_id=model.id.namespace + "/" + model.id.name,
            token=token,
            safe_serialization=True,
            private=True,
        )

        model_id_with_commit = replace(model.id, commit=commit_info.oid)

        # To make sure we get the same hash as validators, we need to redownload it at a
        # local tmp directory after which it can be deleted.
        with tempfile.TemporaryDirectory() as temp_dir:
            model_with_hash = await self.download_model(
                model_id_with_commit, temp_dir, model_constraints
            )
            # Return a ModelId with both the correct commit and hash.
            return model_with_hash.id

    async def download_model(
        self,
        model_id: ModelId,
        local_path: str,
        model_constraints: ModelConstraints,
        max_size_bytes: int = 15 * 1024 * 1024 * 1024,
    ) -> Model:
        """Retrieves a trained model from Hugging Face."""
        if not model_id.commit:
            raise ValueError("No Hugging Face commit id found to read from the hub.")

        repo_id = model_id.namespace + "/" + model_id.name
        token = HuggingFaceModelStore.get_access_token_if_exists()

        # Check ModelInfo for the size of model.safetensors file before downloading.
        api = HfApi()
        model_info = api.model_info(
            repo_id=repo_id,
            revision=model_id.commit,
            timeout=10,
            files_metadata=True,
            token=token,
        )
        size = sum(repo_file.size for repo_file in model_info.siblings)
        if size > max_size_bytes:
            raise ValueError(
                f"Hugging Face repo over maximum size limit. Size {size}. Limit {max_size_bytes}."
            )

        # Transformers library can pick up a model based on the hugging face path (username/model) + rev.
        model = AutoModelForCausalLM.from_pretrained(
            pretrained_model_name_or_path=repo_id,
            revision=model_id.commit,
            cache_dir=local_path,
            use_safetensors=True,
            token=token,
            **model_constraints.kwargs,
        )

        # Get the directory the model was stored to.
        model_dir = utils.get_hf_download_path(local_path, model_id)

        # Realize all symlinks in that directory since Transformers library does not support avoiding symlinks.
        utils.realize_symlinks_in_directory(model_dir)

        # Compute the hash of the downloaded model.
        model_hash = utils.get_hash_of_directory(model_dir)
        model_id_with_hash = replace(model_id, hash=model_hash)

        return Model(id=model_id_with_hash, pt_model=model)



import copy
import pickle
import threading
from typing import Dict, Optional, Set

import bittensor as bt

from taoverse.model.data import ModelMetadata


class ModelTracker:
    """Tracks the current model for each miner.

    Thread safe.
    """

    def __init__(
        self,
    ):
        # Create a dict from miner hotkey to model metadata.
        self.miner_hotkey_to_model_metadata_dict = dict()

        # Make this class thread safe because it will be accessed by multiple threads.
        # One for the downloading new models loop and one for the validating models loop.
        self.lock = threading.RLock()

    def save_state(self, filepath):
        """Save the current state to the provided filepath."""

        # Open a writable binary file for pickle.
        with self.lock:
            with open(filepath, "wb") as f:
                pickle.dump(self.miner_hotkey_to_model_metadata_dict, f)

    def load_state(self, filepath):
        """Load the state from the provided filepath."""

        # Open a readable binary file for pickle.
        with open(filepath, "rb") as f:
            self.miner_hotkey_to_model_metadata_dict = pickle.load(f)

    def get_miner_hotkey_to_model_metadata_dict(self) -> Dict[str, ModelMetadata]:
        """Returns the mapping from miner hotkey to model metadata."""

        # Return a copy to ensure outside code can't modify the scores.
        with self.lock:
            return copy.deepcopy(self.miner_hotkey_to_model_metadata_dict)

    def get_model_metadata_for_miner_hotkey(
        self, hotkey: str
    ) -> Optional[ModelMetadata]:
        """Returns the model metadata for a given hotkey if any."""

        with self.lock:
            if hotkey in self.miner_hotkey_to_model_metadata_dict:
                return self.miner_hotkey_to_model_metadata_dict[hotkey]
            return None

    def on_hotkeys_updated(self, incoming_hotkeys: Set[str]):
        """Notifies the tracker which hotkeys are currently being tracked on the metagraph."""

        with self.lock:
            existing_hotkeys = set(self.miner_hotkey_to_model_metadata_dict.keys())
            for hotkey in existing_hotkeys - incoming_hotkeys:
                del self.miner_hotkey_to_model_metadata_dict[hotkey]
                bt.logging.trace(f"Removed outdated hotkey: {hotkey} from ModelTracker")

    def on_miner_model_updated(
        self,
        hotkey: str,
        model_metadata: ModelMetadata,
    ) -> None:
        """Notifies the tracker that a miner has had their associated model updated.

        Args:
            hotkey (str): The miner's hotkey.
            model_metadata (ModelMetadata): The latest model metadata of the miner.
        """
        with self.lock:
            self.miner_hotkey_to_model_metadata_dict[hotkey] = model_metadata

            bt.logging.trace(f"Updated Miner {hotkey}. ModelMetadata={model_metadata}.")



import dataclasses
from typing import ClassVar, Optional, Type

from transformers import PreTrainedModel

# The maximum bytes for metadata on the chain.
MAX_METADATA_BYTES = 128
# The length, in bytes, of a git commit hash.
GIT_COMMIT_LENGTH = 40
# The length, in bytes, of a base64 encoded sha256 hash.
SHA256_BASE_64_LENGTH = 44
# The max length, in characters, of the competition id
MAX_COMPETITION_ID_LENGTH = 2


@dataclasses.dataclass(frozen=True)
class ModelId:
    """Uniquely identifies a trained model"""

    MAX_REPO_ID_LENGTH: ClassVar[int] = (
        MAX_METADATA_BYTES
        - GIT_COMMIT_LENGTH
        - SHA256_BASE_64_LENGTH
        - MAX_COMPETITION_ID_LENGTH
        - 4  # separators
    )

    # Namespace where the model can be found. ex. Hugging Face username/org.
    namespace: str

    # Name of the model.
    name: str

    # Identifier for competition
    competition_id: int

    # When handling a model locally the commit and hash are not necessary.
    # Commit must be filled when trying to download from a remote store.
    commit: Optional[str] = dataclasses.field(default=None)

    # Hash is filled automatically when uploading to or downloading from a remote store.
    hash: Optional[str] = dataclasses.field(default=None)

    # The secure hash that's used for validation.
    secure_hash: Optional[str] = dataclasses.field(default=None)

    def to_compressed_str(self) -> str:
        """Returns a compressed string representation."""
        return f"{self.namespace}:{self.name}:{self.commit}:{self.secure_hash}:{self.competition_id}"

    @classmethod
    def from_compressed_str(cls, cs: str) -> Type["ModelId"]:
        """Returns an instance of this class from a compressed string representation"""
        tokens = cs.split(":")
        competition_id = int(tokens[4])

        return cls(
            namespace=tokens[0],
            name=tokens[1],
            commit=tokens[2] if tokens[2] != "None" else None,
            secure_hash=tokens[3] if tokens[3] != "None" else None,
            competition_id=competition_id,
        )


@dataclasses.dataclass
class Model:
    """Represents a pre trained foundation model."""

    class Config:
        arbitrary_types_allowed = True

    # Identifier for this model.
    id: ModelId

    # PreTrainedModel.base_model returns torch.nn.Module if needed.
    pt_model: PreTrainedModel


@dataclasses.dataclass
class ModelMetadata:
    # Identifier for this trained model.
    id: ModelId

    # Block on which this model was uploaded on the chain.
    block: int



import unittest
from typing import List, Tuple
from unittest import mock

import bittensor as bt
import torch

from taoverse.metagraph import utils


class TestMetagraphUtils(unittest.TestCase):
    def test_get_high_stake_validators(self):
        # Create a metagraph with 10 neurons of varying stake with the top 4 having a validator permit.
        mock_metagraph = mock.MagicMock()
        mock_metagraph.S = torch.tensor(
            [0, 1, 2, 300, 4, 5, 600, 7, 8, 9], dtype=torch.float32
        )
        mock_metagraph.validator_permit = torch.tensor(
            [
                False,
                False,
                False,
                True,
                False,
                False,
                True,
                False,
                True,
                True,
            ],
            dtype=torch.bool,
        )

        # Check 300 or above stake.
        self.assertEqual(utils.get_high_stake_validators(mock_metagraph, 300), {3, 6})

        # Check 0 or above stake gets only those with permits.
        self.assertEqual(
            utils.get_high_stake_validators(mock_metagraph, 0), {3, 6, 8, 9}
        )

    def _create_metagraph(self):
        """Returns a mocked metagraph with 2 miners and 2 valis."""
        mock_metagraph = mock.MagicMock()
        stakes = torch.tensor([0, 200, 2, 30], dtype=torch.float32)
        mock_metagraph.S = stakes
        mock_metagraph.validator_permit = stakes >= 30
        return mock_metagraph

    def _neuron_info_with_weights(
        self, uid: int, weights: List[Tuple[int, float]]
    ) -> bt.NeuronInfo:
        return bt.NeuronInfo(
            uid=uid,
            netuid=0,
            active=0,
            stake=bt.Balance.from_rao(0),
            stake_dict={},
            total_stake=bt.Balance.from_rao(0),
            rank=0,
            emission=0,
            incentive=0,
            consensus=0,
            trust=0,
            validator_trust=0,
            dividends=0,
            last_update=0,
            validator_permit=False,
            weights=weights,
            bonds=[],
            prometheus_info=None,
            axon_info=None,
            is_null=True,
            coldkey="000000000000000000000000000000000000000000000000",
            hotkey="000000000000000000000000000000000000000000000000",
            pruning_score=0,
        )

    def test_get_top_miners_deduplicated(self):
        """Tests get_top_miners, when validators agree on the top miners."""
        metagraph = self._create_metagraph()

        # Set validator weights such that they agree on miner 0 as the top miner.
        metagraph.neurons = [
            self._neuron_info_with_weights(uid=0, weights=[]),
            self._neuron_info_with_weights(uid=1, weights=[(0, 1)]),
            self._neuron_info_with_weights(uid=2, weights=[]),
            self._neuron_info_with_weights(uid=3, weights=[(0, 1)]),
        ]

        # Verify the miner UID is deduped.
        self.assertSequenceEqual(
            utils.get_top_miners(
                metagraph, min_vali_stake=30, min_miner_weight_percent=0.1
            ),
            {0},
        )

    def test_get_top_miners_multiple_miners(self):
        """Tests get_top_miners, when validators disagree on the top miner."""
        metagraph = self._create_metagraph()

        metagraph.neurons = [
            self._neuron_info_with_weights(uid=0, weights=[]),
            self._neuron_info_with_weights(uid=1, weights=[(0, 1)]),
            self._neuron_info_with_weights(uid=2, weights=[]),
            self._neuron_info_with_weights(uid=3, weights=[(2, 1)]),
        ]
        top_miners = utils.get_top_miners(
            metagraph, min_vali_stake=30, min_miner_weight_percent=0.1
        )
        self.assertEqual(len(top_miners), 2)
        self.assertEqual(top_miners, {0, 2})

    def test_get_top_miners_multiple_weights_set(self):
        """Tests get_top_miners, when validators assign multiple weights"""
        metagraph = self._create_metagraph()

        # Have vali 1 set multiple weights, ensuring some are above and below the threshold.
        # Note that although uid 1 has 0.1 weight, this is only 5% of the weight.
        metagraph.neurons = [
            self._neuron_info_with_weights(uid=0, weights=[]),
            self._neuron_info_with_weights(uid=1, weights=[(0, 1), (1, 0.1), (2, 0.5)]),
            self._neuron_info_with_weights(uid=2, weights=[]),
            self._neuron_info_with_weights(uid=3, weights=[]),
        ]
        self.assertEqual(
            utils.get_top_miners(
                metagraph, min_vali_stake=30, min_miner_weight_percent=0.1
            ),
            {0, 2},
        )


if __name__ == "__main__":
    unittest.main()



import concurrent
from datetime import datetime, timedelta
import functools
import multiprocessing
import os
import random
from typing import Any, Optional

import bittensor as bt


def _wrapped_func(func: functools.partial, queue: multiprocessing.Queue):
    try:
        result = func()
        queue.put(result)
    except (Exception, BaseException) as e:
        # Catch exceptions here to add them to the queue.
        queue.put(e)


def run_in_subprocess(func: functools.partial, ttl: int, mode="fork") -> Any:
    """Runs the provided function on a subprocess with 'ttl' seconds to complete.

    Args:
        func (functools.partial): Function to be run.
        ttl (int): How long to try for in seconds.
        mode (str): Mode by which the multiprocessing context is obtained. Default to fork for pickling.

    Returns:
        Any: The value returned by 'func'
    """
    ctx = multiprocessing.get_context(mode)
    queue = ctx.Queue()
    process = ctx.Process(target=_wrapped_func, args=[func, queue])

    process.start()

    process.join(timeout=ttl)

    if process.is_alive():
        process.terminate()
        process.join()
        raise TimeoutError(f"Failed to {func.func.__name__} after {ttl} seconds")

    # Raises an error if the queue is empty. This is fine. It means our subprocess timed out.
    result = queue.get(block=False)

    # If we put an exception on the queue then raise instead of returning.
    if isinstance(result, Exception):
        raise result
    if isinstance(result, BaseException):
        raise Exception(f"BaseException raised in subprocess: {str(result)}")

    return result


def run_in_thread(func: functools.partial, ttl: int, name=None) -> Any:
    """Runs the provided function on a thread with 'ttl' seconds to complete.

    Args:
        func (functools.partial): Function to be run.
        ttl (int): How long to try for in seconds.

    Returns:
        Any: The value returned by 'func'
    """

    executor = concurrent.futures.ThreadPoolExecutor(max_workers=1)

    try:
        future = executor.submit(func)
        return future.result(timeout=ttl)
    except concurrent.futures.TimeoutError as e:
        bt.logging.error(f"Failed to complete '{name}' within {ttl} seconds.")
        raise TimeoutError(f"Failed to complete '{name}' within {ttl} seconds.") from e
    finally:
        bt.logging.trace(f"Completed {name}")
        executor.shutdown(wait=False)
        bt.logging.trace(f"{name} cleaned up successfully")


def get_version(filepath: str) -> Optional[int]:
    """Loads a version from the provided filepath or None if the file does not exist.

    Args:
        filepath (str): Path to the version file."""
    if os.path.exists(filepath):
        with open(filepath, "r") as f:
            line = f.readline()
            if line:
                return int(line)
            return None
    return None


def save_version(filepath: str, version: int):
    """Saves a version to the provided filepath."""
    os.makedirs(os.path.dirname(filepath), exist_ok=True)
    with open(filepath, "w") as f:
        f.write(str(version))


def random_date(start: datetime, end: datetime, seed: int = None) -> datetime:
    """Return a random datetime between two datetimes.

    Args:
        start (datetime): Start of the range, inclusive.
        end (datetime): End of the range, inclusive.
        seed (int): Optional Seed for the random number generator.
    """

    if start.tzinfo != end.tzinfo:
        raise ValueError("Start and end must have the same timezone.")

    if start >= end:
        raise ValueError("Start must be before end.")

    if seed:
        random.seed(seed)

    # Choose a random point between the 2 datetimes.
    random_seconds = random.randint(0, int((end - start).total_seconds()))

    # Add the random seconds to the start time
    return start + timedelta(seconds=random_seconds)



import abc

from taoverse.model.competition.data import Competition
from taoverse.model.data import Model, ModelId


class RemoteModelStore(abc.ABC):
    """An abstract base class for storing and retrieving a pre trained model."""

    @abc.abstractmethod
    async def upload_model(self, model: Model, competition: Competition) -> ModelId:
        """Uploads a trained model in the appropriate location based on implementation."""
        pass

    @abc.abstractmethod
    async def download_model(
        self, model_id: ModelId, local_path: str, competition: Competition
    ) -> Model:
        """Retrieves a trained model from the appropriate location and stores at the given path."""
        pass



from dataclasses import dataclass, field
from typing import Any, List, Type, Optional

from transformers import PreTrainedModel


@dataclass
class NormValidationConstraints:
    "Defines the validation constriants to be checked based on model norms."

    norm_eps_soft: int
    norm_eps_soft_percent_threshold: float
    norm_eps_hard: int


@dataclass
class ModelConstraints:
    """Defines the constraints for models submitted to a specific competition."""

    # The maximum parameter size allowed for models
    max_model_parameter_size: int

    # Architecture class of model
    allowed_architectures: List[Type[PreTrainedModel]]

    # The model's sequence length.
    sequence_length: int

    # The Pretrained tokenizer to use.
    tokenizer: str

    # Block delay before evaluating uploaded models. Based on look-back period for eval data collection.
    eval_block_delay: int

    # Any additional arguments to pass to from_pretrained
    kwargs: Any = field(default_factory=dict)

    # Norm validation values.
    norm_validation_constraints: Optional[NormValidationConstraints] = None


@dataclass
class Competition:
    """Defines a competition."""

    # Unique ID for this competition.
    # Recommend making an IntEnum for use in the subnet codebase.
    id: int

    # All restrictions on models allowed in this competition.
    constraints: ModelConstraints

    # Percentage of emissions dedicated to this competition.
    reward_percentage: float



import bisect
import copy
import random
import threading
from typing import List


class MinerIterator:
    """A thread safe infinite iterator to cyclically enumerate the current set of miner UIDs.

    Why? To perform miner evaluations, the validator will enumerate through the miners in order to help ensure
    each miner is evaluated at least once per epoch.
    """

    def __init__(self, miner_uids: List[int]):
        self.miner_uids = sorted(copy.deepcopy(miner_uids))
        # Start the index at a random position. This helps ensure that miners with high UIDs aren't penalized if
        # the validator restarts frequently.
        self.index = random.randint(0, len(self.miner_uids) - 1)
        self.lock = threading.Lock()

    def __iter__(self):
        return self

    def __next__(self) -> int:
        with self.lock:
            if len(self.miner_uids) == 0:
                # This iterator should be infinite. If there are no miner UIDs, raise an error.
                raise IndexError("No miner UIDs.")

            uid = self.miner_uids[self.index]
            self.index += 1
            if self.index >= len(self.miner_uids):
                self.index = 0
            return uid

    def peek(self) -> int:
        """Returns the next miner UID without advancing the iterator."""
        with self.lock:
            if len(self.miner_uids) == 0:
                # This iterator should be infinite. If there are no miner UIDs, raise an error.
                raise IndexError("No miner UIDs.")

            return self.miner_uids[self.index]

    def set_miner_uids(self, miner_uids: List[int]):
        """Updates the miner UIDs to iterate.

        The iterator will be updated to the first miner uid that is greater than or equal to UID that would be next
        returned by the iterator. This helps ensure that frequent updates to the miner_uids does not cause too much
        churn in the sequence of UIDs returned by the iterator.
        """
        sorted_uids = sorted(copy.deepcopy(miner_uids))
        with self.lock:
            next_uid = self.miner_uids[self.index]
            new_index = bisect.bisect_left(sorted_uids, next_uid)
            if new_index >= len(sorted_uids):
                new_index = 0
            self.index = new_index
            self.miner_uids = sorted_uids



import statistics
from typing import List, Optional, Tuple, Dict

import bittensor as bt

from taoverse.model.competition import utils as competition_utils
from taoverse.model.competition.data import Competition, ModelConstraints
from taoverse.model.data import Model, ModelMetadata
from taoverse.model.model_tracker import ModelTracker
from taoverse.model.storage.local_model_store import LocalModelStore
from taoverse.model.storage.model_metadata_store import ModelMetadataStore
from taoverse.model.storage.remote_model_store import RemoteModelStore
from taoverse.model.utils import get_hash_of_two_strings


class MinerMisconfiguredError(Exception):
    """Error raised when a miner is misconfigured."""

    def __init__(self, hotkey: str, message: str):
        self.hotkey = hotkey
        super().__init__(f"[{hotkey}] {message}")


class ModelUpdater:
    """Checks if the currently tracked model for a hotkey matches what the miner committed to the chain."""

    def __init__(
        self,
        metadata_store: ModelMetadataStore,
        remote_store: RemoteModelStore,
        local_store: LocalModelStore,
        model_tracker: ModelTracker,
    ):
        self.metadata_store = metadata_store
        self.remote_store = remote_store
        self.local_store = local_store
        self.model_tracker = model_tracker

    @staticmethod
    def verify_model_satisfies_parameters(
        model: Model, model_constraints: ModelConstraints
    ) -> bool:
        if not model_constraints:
            bt.logging.trace(f"No competition found for {model.id.competition_id}")
            return False

        # Check that the parameter count of the model is within allowed bounds.
        parameter_size = sum(p.numel() for p in model.pt_model.parameters())
        if parameter_size > model_constraints.max_model_parameter_size:
            return False

        # Make sure it's an allowed architecture.
        if type(model.pt_model) not in model_constraints.allowed_architectures:
            return False

        # Check parameters are sane if specified
        if model_constraints.norm_validation_constraints is not None:
            return ModelUpdater._validate_parameters(
                model.pt_model,
                model_constraints.norm_validation_constraints.norm_eps_soft,
                model_constraints.norm_validation_constraints.norm_eps_soft_percent_threshold,
                model_constraints.norm_validation_constraints.norm_eps_hard,
            )

        return True

    async def _get_metadata(self, hotkey: str) -> Optional[ModelMetadata]:
        """Get metadata about a model by hotkey"""
        return await self.metadata_store.retrieve_model_metadata(hotkey)

    async def sync_model(
        self,
        hotkey: str,
        curr_block: int,
        schedule_by_block: List[Tuple[int, List[Competition]]],
        force: bool = False,
    ) -> bool:
        """Updates local model for a hotkey if out of sync and returns if it was updated."

        Args:
           hotkey (str): The hotkey of the model to sync.
           curr_block (int): The current block.
           force (bool): Whether to force a sync for this model, even if it's chain metadata hasn't changed.
           schedule_by_block (List[Tuple[int, List[Competition]]]): Which competitions are being run at a given block.
        """
        # Get the metadata for the miner.
        metadata = await self._get_metadata(hotkey)

        if not metadata:
            raise MinerMisconfiguredError(
                hotkey, f"No valid metadata found on the chain"
            )

        # Check that the metadata indicates a competition available at time of upload.
        competition = competition_utils.get_competition_for_block(
            comp_id=metadata.id.competition_id,
            block=metadata.block,
            schedule_by_block=schedule_by_block,
        )
        if not competition:
            raise MinerMisconfiguredError(
                hotkey,
                f"No competition found for {metadata.id.competition_id} at block {metadata.block}",
            )

        # Check that the metadata is old enough to meet the eval_block_delay for the competition.
        # If not we return false and will check again next time we go through the update loop.
        if curr_block - metadata.block < competition.constraints.eval_block_delay:
            bt.logging.debug(
                f"""Sync for hotkey {hotkey} delayed as the current block: {curr_block} is not at least 
                {competition.constraints.eval_block_delay} blocks after the upload block: {metadata.block}. 
                Will automatically retry later."""
            )
            return False

        # Check what model id the model tracker currently has for this hotkey.
        tracker_model_metadata = self.model_tracker.get_model_metadata_for_miner_hotkey(
            hotkey
        )
        # If we are not forcing a sync due to retrying a top model we can short-circuit if no change.
        if not force and metadata == tracker_model_metadata:
            return False

        # Get the local path based on the local store to download to (top level hotkey path)
        path = self.local_store.get_path(hotkey)

        # Otherwise we need to download the new model based on the metadata.
        model = await self.remote_store.download_model(
            metadata.id, path, competition.constraints
        )

        # Update the tracker even if the model fails the following checks to avoid redownloading without new metadata.
        self.model_tracker.on_miner_model_updated(hotkey, metadata)

        # Check that the hash of the downloaded content matches.
        secure_hash = get_hash_of_two_strings(model.id.hash, hotkey)
        if secure_hash != metadata.id.secure_hash:
            raise MinerMisconfiguredError(
                hotkey,
                f"Hash of content downloaded from hugging face does not match chain metadata. {metadata}",
            )

        if not ModelUpdater.verify_model_satisfies_parameters(
            model, competition.constraints
        ):
            raise MinerMisconfiguredError(
                hotkey,
                f"Model does not satisfy parameters for competition {competition.id}",
            )

        return True

    @staticmethod
    def _validate_parameters(
        base_model, eps_soft, eps_soft_percent_threshold, eps_hard, print_vals=False
    ) -> bool:
        """
        Validate that parameters of a model

        Parameters:
            base_model (transformers.PreTrainedModel): The base model instance.
            num_layers (int): Number of layers in the model to inspect.
            eps_soft (float): Calculate the percentage of layers above this norm
            eps_soft_percent_threshold (float): Threshold of percentage above eps_soft that will trigger a detection
            eps_hard (float): Hard limit for any norm
        """

        exceed_counts = {
            "q_proj": 0,
            "k_proj": 0,
            "v_proj": 0,
            "o_proj": 0,
            "up_proj": 0,
            "down_proj": 0,
        }
        total_counts = {
            "q_proj": 0,
            "k_proj": 0,
            "v_proj": 0,
            "o_proj": 0,
            "up_proj": 0,
            "down_proj": 0,
        }
        if print_vals:
            avg_norms = {
                "q_proj": 0.0,
                "k_proj": 0.0,
                "v_proj": 0.0,
                "o_proj": 0.0,
                "up_proj": 0.0,
                "down_proj": 0.0,
            }
            max_norms = {
                "q_proj": 0.0,
                "k_proj": 0.0,
                "v_proj": 0.0,
                "o_proj": 0.0,
                "up_proj": 0.0,
                "down_proj": 0.0,
            }

        for layer in base_model.model.layers:
            for proj in ["q_proj", "k_proj", "v_proj", "o_proj"]:
                weight_norm = getattr(layer.self_attn, proj).weight.norm().item()
                if weight_norm > eps_hard:
                    return False
                elif weight_norm > eps_soft:
                    exceed_counts[proj] += 1
                total_counts[proj] += 1
                if print_vals:
                    avg_norms[proj] += weight_norm
                    max_norms[proj] = max(max_norms[proj], weight_norm)

            # up_proj and down_proj are in the mlp layer
            for proj in ["up_proj", "down_proj"]:
                weight_norm = getattr(layer.mlp, proj).weight.norm().item()
                if weight_norm > eps_hard:
                    return False
                elif weight_norm > eps_soft:
                    exceed_counts[proj] += 1
                total_counts[proj] += 1
                if print_vals:
                    avg_norms[proj] += weight_norm
                    max_norms[proj] = max(max_norms[proj], weight_norm)

        # Calculating and printing percentages
        percentages = [
            exceed_counts[proj] / total_counts[proj] for proj in exceed_counts
        ]

        if print_vals:
            for key, value in total_counts.items():
                avg_norms[key] = avg_norms[key] / value
            print(avg_norms)
            print(max_norms)
            print(percentages)

        return statistics.fmean(percentages) <= eps_soft_percent_threshold



import argparse
import enum


class IntEnumAction(argparse.Action):
    """
    Argparse action for handling Enums
    """

    def __init__(self, **kwargs):
        # Pop off the type value
        enum_type = kwargs.pop("type", None)

        # Ensure an Enum subclass is provided
        if enum_type is None:
            raise ValueError("type must be assigned an Enum when using EnumAction")
        if not issubclass(enum_type, enum.Enum):
            raise TypeError("type must be an Enum when using EnumAction")

        # Generate choices from the Enum
        kwargs.setdefault("choices", tuple(str(e.value) for e in enum_type))

        super(IntEnumAction, self).__init__(**kwargs)

        self._enum = enum_type

    def __call__(self, parser, namespace, values, option_string=None):
        # Convert value back into an Enum
        value = self._enum(int(values))
        setattr(namespace, self.dest, value)



import time

import numpy as np


class PerfSample:
    def __init__(self, perf_tracker):
        self.perf_tracker = perf_tracker
        self.start_time = None

    def __enter__(self):
        self.start_time = time.monotonic_ns()
        return self

    def __exit__(self, exception_type, exception_value, exception_traceback):
        duration = time.monotonic_ns() - self.start_time
        self.perf_tracker.samples.append(duration)


class PerfMonitor:
    """PerfMonitor is a context manager that tracks the performance of a block of code by taking several samples.

    Example:
        tracker = PerfMonitor("MyOperation")
        for _ in range(10):
            with tracker.sample():
                // Do something

        print(tracker.summary_str())
    """

    def __init__(self, name):
        self.name = name
        self.samples = []

    def set_samples_for_testing(self, samples):
        self.samples = samples

    def sample(self) -> PerfSample:
        """Returns a context manager that will record the duration of the block it wraps."""
        return PerfSample(self)

    def min(self) -> float:
        """Returns the minimum duration recorded by the tracker in seconds."""
        return self._ns_to_s(np.min(self.samples))

    def max(self) -> float:
        """Returns the maximum duration recorded by the tracker in seconds."""
        return self._ns_to_s(np.max(self.samples))

    def median(self) -> float:
        """Returns the median duration recorded by the tracker in seconds."""
        return self._ns_to_s(np.median(self.samples))

    def percentile(self, p: float) -> float:
        """Returns the pth percentile duration recorded by the tracker in seconds."""
        return self._ns_to_s(np.percentile(self.samples, p))

    def summary_str(self) -> str:
        """Returns a string summarizing the performance of the tracked operation."""
        if not self.samples:
            return f"{self.name} performance: N=0"

        durations_ns = np.array(self.samples)

        return (
            f"{self.name} performance: N={len(durations_ns)} | "
            + f"Min={self._format_duration(np.min(durations_ns))} | "
            + f"Max={self._format_duration(np.max(durations_ns))} | "
            + f"Median={self._format_duration(np.median(durations_ns))} | "
            + f"P90={self._format_duration(np.percentile(durations_ns, 90))}"
        )

    def _format_duration(self, duration_ns: int) -> str:
        units = [
            ("ns", 1),
            ("μs", 1000),
            ("ms", 1000_000),
            ("s", 1000_000_000),
            ("min", 60 * 1000_000_000),
        ]

        for unit, divisor in reversed(units):
            if duration_ns >= divisor:
                return f"{duration_ns/divisor:.2f} {unit}"

        return f"{duration_ns:.2f} ns"

    def _ns_to_s(self, duration_ns: int) -> float:
        return duration_ns / 1_000_000_000
